// 
// deferred shading fx
//
// This shader assumes that the scene geometry has been rendered into multiple render targets
// for diffuse, normal, and depth information
//
// The lighting is done in world space,
// many examples of deferred shading do this in view/camera space.
//
//
// Global variables
//--------------------------------------------------------------------------------------


float3 lightposn={3.0,4.0,5.0};  // in world space
float  lightradius=10.0f;
float3 lightcolor = {0.6,0.6,0.6};
float  gloss = 0.5;   // affects how much specular is reflected.  // should be a property of the material, todo: put into alpha on diffuse.
float4x4 ViewInv: ViewInverse;
float4x4 Proj;

texture diffusetex,normaltex,positiontex; 
sampler sDiffuse  = sampler_state 
{ 
	texture = <diffusetex>;  
	AddressU  = CLAMP;        
	AddressV  = CLAMP;
	MINFILTER = POINT;
	MAGFILTER = POINT;
	MIPFILTER = NONE;
};

sampler sNormals  = sampler_state 
{ 
	texture = <normaltex>;  
	AddressU  = CLAMP;        
	AddressV  = CLAMP;
	MINFILTER = POINT;
	MAGFILTER = POINT;
	MIPFILTER = NONE;
};

sampler sPosition = sampler_state 
{
	texture = <positiontex>; 
	AddressU  = CLAMP;        
	AddressV  = CLAMP;
	MINFILTER = POINT;
	MAGFILTER = POINT;
	MIPFILTER = NONE;
};

float oneoverwidth=1.0f/640.0f; 
float oneoverheight=1.0f/480.0f;

struct VS_INPUT 
{
   float3 pos:     POSITION;
   float2 tex:     TEXCOORD0; 
};

struct VS_OUTPUT 
{
   float4 pos:  POSITION;
   float2 tex:  TEXCOORD0; 
   float3 vdir: TEXCOORD1;
};

VS_OUTPUT vertex_shader_deferred_pass( VS_INPUT In )
{
	VS_OUTPUT Out;

	Out.pos      = float4(sign(In.pos.xy),0.0,1.0);
	Out.tex.x    = (Out.pos.x+1.0) * 0.5  + oneoverwidth*0.5;
	Out.tex.y    = 1.0 - ((Out.pos.y+1.0) * 0.5) + oneoverheight*0.5f;

	// the following was successfully used to get the world space viewing direction.  
	// The pixel shader can use this with depth info to derive pos instead of storing entire xyz pos.
	float4 camvdir;  // viewing direction in camera space
	camvdir.x   = In.pos.x / Proj[0].x;  // going backward from projected clip space coordinates
	camvdir.y   = In.pos.y / Proj[1].y;
	camvdir.z   = -1.0f;
	camvdir.w   = 0.0f;
	float4 vdir = mul(camvdir,ViewInv);
	Out.vdir    = vdir.xyz;  // viewing direction in world space 

	return Out;
}


float4 deferredLighting_Ps(float2 uv : TexCoord0    , float3 vdir:TEXCOORD1
                          ) : Color0 
{
	// derive world space position based on center of projection, depth at pixel, and viewing direction at this pixel in world space
	float3 cop  = ViewInv[3];
	float depth = tex2D(sPosition, uv).x;  // render target just stores depth info now, not the entire xyz position
	float3 pos = cop + normalize(vdir)*-depth;  // this works since the "depth" is actually distance from viewpoint as in length(xyz) instead of just z
	
	float3 n = (tex2D(sNormals, uv).xyz-float3(0.5,0.5,0.5))*2; // unpack from rendertarget color texture
	float4 diffuse = tex2D(sDiffuse, uv);
	float3 e = normalize(ViewInv[3] - pos);  //float4x4 matVi : ViewInverse;
	float3 lightvec = lightposn - pos;
	float  falloff = saturate(1-length(lightvec)/lightradius);
	float3 l = normalize(lightvec); //float3 lightDir = normalize(lightvec);
	float3 h = normalize(e+l);
	float spec = pow(saturate(dot(n,h)),64);  // specular intensity
	float d = saturate(dot(n,l));
	float4 color;
	color.xyz = ((lightcolor * d)  *  diffuse + lightcolor * gloss * spec) * falloff ; 
	color.w = 1.0;
	//color.xyz = tex2D(sNormals, uv);
	return color;
}

technique t0
{
	pass p0
	{
		DEPTHBIAS = 0;
		STENCILENABLE=    TRUE  ;
		TWOSIDEDSTENCILMODE = FALSE;
		CCW_STENCILPASS=KEEP;
		STENCILPASS=     KEEP;
		STENCILFAIL=     KEEP;
		STENCILZFAIL=     KEEP;
		ZWRITEENABLE=     FALSE ;
		ZENABLE = TRUE;
		ZFUNC = LESSEQUAL ;
		STENCILREF=  0x01 ;
		STENCILFUNC= GREATER; // since unlike shadowpoly we want to draw where shadow isn't  // LESSEQUAL ;
		FOGENABLE=        FALSE ;
		ALPHABLENDENABLE= TRUE  ;
		SHADEMODE = GOURAUD ;
		SRCBLEND=  ONE ;
		DESTBLEND= ONE ;
		CullMode = CW;
		FILLMODE=SOLID;
		// VertexShader = null;
		VertexShader = compile vs_2_0 vertex_shader_deferred_pass();
		PixelShader = compile ps_2_0 deferredLighting_Ps();
	}
}
